#include <iostream.h>
#include <hpldbin.h>
#include <dpldbin.h>
#include <popsmpl.h>
// #include <popelite.h>
// #include <popcrd.h>
// #include <slctrw.h>
// #include <slctrwwr.h>
// #include <slcttour.h>
// #include <slctlrnk.h>
#include <slctsus.h>
#include <schemata.h>
#include <ga.h>
#include <Normal.h>
#include <GetOpt.h>

ParameterMap params1(16,-5.12,5.12);
ParameterMap params2(16,-2.048,2.048);
ParameterMap params3(16,-1.28,1.28);
ParameterMap params4(16,-65.536,65.536);
Normal  gauss_0_1(0,1,&mlcg); 

unsigned nGen = 400 , uPopSize = 30, uTest = 1;
float    flMRate = 0.003;
float    flXRate = 1;                              
float    flGap   = 0.1;                      
SigmaScaling<> *pScale = new SigmaScaling<>(1.5);
// ScalingScheme<> *pScale = new PowerScaling<>(1.2);


//                              
//      F1 : f1(x1, x2, x3) = x1 ^ 2 + x2 ^ 2 + x3 ^ 2
//                  
//              -5.12 <= xi <= 5.12  
//                      
//      use 16 bits for each parameter
//                                    
//      
class F1 
{ 
public:
	double operator()(const TGAObj * obj) 
        {
	   double x1 = params1((BinIndividual *) obj, 0);
	   double x2 = params1((BinIndividual *) obj, 1);
	   double x3 = params1((BinIndividual *) obj, 2);

	   return x1 * x1 + x2 * x2 + x3 * x3;
        }
};

//                              2                 2
//      F2 : f2(x1, x2) = 100(x1  - x2) + (1 - x1)    
//                  
//              -2.048 <= xi <= 2.048
//                      
//      
class F2 
{ 
public:
	double operator()(const TGAObj * obj)
        {
	    double x1 = params2((BinIndividual *) obj, 0);
	    double x2 = params2((BinIndividual *) obj, 1);

            return 100 * pow(x1 * x1 - x2, 2.0) + pow(1 - x1, 2.0);
        }
};

//
//      F3 : f3(x1, x2, x3, x4, x5) =              
//              30 + floor(x1) + floor(x2) + floor(x3) + floor(x4) + floor(x5)  
//                  
//              -5.12 <= xi <= 5.12  
//                      
class F3
{ 
public:
	double operator()(const TGAObj * obj) 
        {
            double      flValues[5];
            int         nValues[5], i, retval;
                
            for (retval = 0, i = 0; i < 5; i++) {
                flValues[i] =  params1((BinIndividual*) obj,i);
                nValues[i] = (int) flValues[i];
                if (nValues[i] > flValues[i])                              
                    nValues[i] -= 1;
                retval += nValues[i];
            }
            return 30 + retval;
        }
};

//                             30     
//                           ------
//      F4 : f4(xi) =        \       [ i * (xi)**4  + Gauss(0,1) ]
//                           /
//                           ------
//                             1
//
//              -1.28 <= xi <= 1.28    
//                                                                              
class F4
{ 
public:
	double operator()(const TGAObj * obj) 
        {
             double   flTemp, flTotal = 0;
             for(int i=0, j=0; i < 30; i++, j+=8) {           
                 flTemp =  params3((BinIndividual*) obj,j);
                 flTemp = i * pow(flTemp, 4.0);
                 flTemp += gauss_0_1();
                 flTotal += flTemp;
             }

             return flTotal;
        }
};



//      f5(x1,x2) [Shekel's foxholes] =
//
//
//                             25
//                           ------                1
//             0.002  +      \       ------------------------------
//                           /                     2
//                           ------               ----              6
//                             j=1        j  +    \     (x   - a   )
//                                                /___    i     ij
//                                                i=1                           
//                                       
//              -65.536 <= xi <= 65.536
//                                                  
//                     
class F5
{ 
public:
	double operator()(const TGAObj * obj)
        {
            int i, j;             
            double ans = 0;
            double fj;
            double x[2];             

            x[0] = params4((BinIndividual *) obj, 0);
            x[1] = params4((BinIndividual *) obj, 1);

            for (j = 0; j < 25; j++)
            {                                                         
                fj = j + 1;                                                     
                for (i = 0; i < 2; i++)  
                {                      
                     fj += pow(x[i] - a[i][j], 6.0);
                }                     
                ans += 1.0 / fj;                     
            }                                               
            return(0.002 + ans);
        }
private:
static int a[2][25];
};

int F5::a[2][25] ={
  {-32, -16, 0, 16, 32, -32, -16, 0, 16, 32, -32, -16, 0, 16, 32,
   -32, -16, 0, 16, 32, -32, -16, 0, 16, 32        },
  {-32, -32, -32, -32, -32, -16, -16, -16, -16, -16,
   16, 16, 16, 16, 16, 32, 32, 32, 32, 32  }
};

void    printBanner(const char *prog)
{
           cout << "usage: " << prog << " [-t test] [-n gen] [-p popsize] [-x xp
t] [-m mut] [-g gap]" << endl;
           cout << "-t test     select one of the five dejong's test functions"
<< endl;
           cout << "-n gen      the number of generations to perform" << endl;
           cout << "-p popsize  the population size" << endl;
           cout << "-x xrate    the crossover rate" << endl;
           cout << "-m mrate    the mutation rate" << endl;
           cout << "-g gap      the generation gap" << endl;
}

class HammingDistance 
{
public:
	double operator()(const TGAObj* obj1, const TGAObj* obj2) const
        {
		return ((const BinIndividual *) obj1)->hammingDistance(* (const BinIndividual *) obj2);
        }
};

void    runtest(PopulationBase<>* pPop, int nLen)
{
        int i;

        for (i=0; i<uPopSize; i++)
             pPop->add(new BinHaploid(nLen,true));
             // pPop->add(new BinDiploid(nLen,true));

        // SelectionScheme<>* pSelect = new LinearRanking<vector<TGAObj*> >();
        SelectionScheme<>* pSelect = new SUS_Select<vector<TGAObj*> >();
        Crossover*  pXover = new MultiPtCrossover<BinIndividual>(3, flXRate);
       // Crossover* pXover = new BinOnePtCrossover(flXRate); 
        GeneticAlgorithm<>  ga(pPop,
                              pSelect, pXover,                               
                              flMRate,                                         
                              flGap, pScale);
             
        cout << "Generation " << ga.trials() << endl;
        cout << "Best Ind" << endl << "-----------" << endl;
        cout << *ga.bestInd() << endl;
        cout << "Worst Ind" << endl << "-----------" << endl;
        cout << *ga.worstInd() << endl;

        while (ga.trials() < nGen) 
             ga.evolve();

        cout << "Final population" << endl;
        cout << "----------------" << endl;
	ga.printOn(cout);
        cout << endl;
        cout << "Selection performed : " << ga.numSelected() << endl;
        cout << "Crossover performed : " << ga.numCrossed() << endl;
        cout << "Best Individual : " ;
        ga.bestInd()->printOn(cout);
        cout << endl;
        cout << "Worst Individual : " ;
        ga.worstInd()->printOn(cout);
        cout << endl;
        cout << "Online : " << ga.objOnLine() << endl;
        cout << "Offline : " << ga.objOffLine() << endl;

}

void    main(int argc, char *argv[])                   
{
        GetOpt options(argc,argv,"t:n:p:x:m:g:");
        char opt;
                    
        if (argc == 1) {   
           printBanner(argv[0]);                     
           exit(-1);
        }
        else {                                           
           opt = options();                             
           if (opt < 0 || opt == EOF) {
               printBanner(argv[0]);
               exit(-1);                                                       
           }
           else      
             do {
                 switch (opt) {
                 case 't' :
                    uTest = atoi(options.optarg);
                    if ((uTest < 1) || (uTest > 5))
                        uTest = 1;
                    break;
                 case 'n' :
                    nGen = abs(atoi(options.optarg));
                    break;
                 case 'p' :
                    uPopSize = atoi(options.optarg);
                    break;
                 case 'x' :
                    flXRate = fabs(atof(options.optarg));
                    if (flXRate > 1)
                        flXRate = 1;
                    break;
                 case 'm' :
                    flMRate = fabs(atof(options.optarg));
                    if (flMRate > 1)
                        flMRate = 1;
                    break;
                 case 'g' :
                    flGap = fabs(atof(options.optarg));
                    if (flGap > 1)
                        flGap = 1;
                    break;
                 case '?' :
                    printBanner(argv[0]);
                    exit(-1);
                 }
             } while ((opt = options()) != EOF);
        }

        cout << "Test            : " << uTest << endl;
        cout << "Generations     : " << nGen << endl;
        cout << "Population size : " << uPopSize << endl;
        cout << "Crossover rate  : " << flXRate << endl;
        cout << "Mutation rate   : " << flMRate << endl;
        cout << "Generation gap  : " << flGap << endl << endl;

        switch (uTest) {
                case 2: 
                   {
                    SimplePopulation<F2>* pPop = new SimplePopulation<F2>();
                    runtest(pPop, 32);
                   }
                    break;
                case 3:
                    {
                    SimplePopulation<F3>* pPop = new SimplePopulation<F3>();
                    runtest(pPop, 80);
                    }
                    break;
                case 4:
                    {
                    SimplePopulation<F4>* pPop = new SimplePopulation<F4>();
                    runtest(pPop, 480);
                    }
                    break;
                case 5:
                    {
                    SimplePopulation<F5>* pPop = new SimplePopulation<F5>();
                    runtest(pPop, 32);
                    }
                    break;
                default:
                    {
                    SimplePopulation<F1>* pPop = new SimplePopulation<F1>();
                    runtest(pPop, 48);

                    // illustrate the use of class SchemataStatistics
                    // to count the number of schemata in a population
                    //
                    SchemataStatistics<> st;
                    // add three schemata, # is the don't care symbol
                    st.add("1111111111######################################");
                    st.add("1###############################################");
                    st.add("0###############################################");
                    // now count the individuals in the population
                    // having these schemata
                    st(*pPop);
                    st.printOn(cout);
                    cout << endl;
                    }
        }
}


