//
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93, 1996.
//
//      Description
//      -----------
//      This program uses the GENITOR genetic algorithm to evolve
//      a bitstring with all ones
//      
//      execute the program with no arguments to see the options available
//

#include <iostream.h>
#include <hpldbin.h>
#include <popsmpl.h>
#include <slctsus.h>
#include <genitor.h>
#include <GetOpt.h>

unsigned nGen = 400 , uPopSize = 30;
float    flMRate = 0.003;
float    flXRate = 1;                              
int      nLen = 16;

//                                    
// count the number of ones in a bitstring                                    
//      
class CountOnes
{ 
public:
	double operator()(const TGAObj * obj) 
        {
	   return ((const BinIndividual *) obj)->asBitString().count();
        }
};

void    printBanner(const char *prog)
{
        cout << "This program evolves a bit string with all one's" << endl;
        cout << "usage: " << prog << " [-l len] [-n gen] [-p popsize] [-x xp t] [-m mut]" << endl;
        cout << "-l len      the length of each bitstring" << endl;
        cout << "-n gen      the number of generations to perform" << endl;
        cout << "-p popsize  the population size" << endl;
        cout << "-x xrate    the crossover rate" << endl;
        cout << "-m mrate    the mutation rate" << endl;
}

class HammingDistance 
{
public:
	double operator()(const TGAObj* obj1, const TGAObj* obj2) const
        {
		return ((const BinIndividual *) obj1)->hammingDistance(* (const BinIndividual *) obj2);
        }
};

void    main(int argc, char *argv[])                   
{
        GetOpt options(argc,argv,"l:n:p:x:m:g:");
        char opt;
                    
        if (argc == 1) {   
           printBanner(argv[0]);                     
           exit(-1);
        }
        else {                                           
           opt = options();                             
           if (opt < 0 || opt == EOF) {
               printBanner(argv[0]);
               exit(-1);                                                       
           }
           else      
             do {
                 switch (opt) {
                 case 'l' :
                    nLen = atoi(options.optarg);
                    if (nLen < 0) 
                        nLen = 16;
                    break;
                 case 'n' :
                    nGen = abs(atoi(options.optarg));
                    break;
                 case 'p' :
                    uPopSize = atoi(options.optarg);
                    break;
                 case 'x' :
                    flXRate = fabs(atof(options.optarg));
                    if (flXRate > 1)
                        flXRate = 1;
                    break;
                 case 'm' :
                    flMRate = fabs(atof(options.optarg));
                    if (flMRate > 1)
                        flMRate = 1;
                    break;
                 case '?' :
                    printBanner(argv[0]);
                    exit(-1);
                 }
             } while ((opt = options()) != EOF);
        }

        cout << "Individual Length : " << nLen << endl;
        cout << "Generations       : " << nGen << endl;
        cout << "Population size   : " << uPopSize << endl;
        cout << "Crossover rate    : " << flXRate << endl;
        cout << "Mutation rate     : " << flMRate << endl;

        GENITOR<CountOnes>  ga(new BinHaploid(nLen), uPopSize, 
                               new BinTwoPtCrossover(flXRate),
                               flMRate);
                              
        while (ga.trials() < nGen) 
             ga.evolve();

        cout << "Final population" << endl;
        cout << "----------------" << endl;
	ga.printOn(cout);
        cout << endl;
        cout << "Selection performed : " << ga.numSelected() << endl;
        cout << "Crossover performed : " << ga.numCrossed() << endl;
        cout << "Best Individual : " ;
        ga.bestInd()->printOn(cout);
        cout << endl;
        cout << "Worst Individual : " ;
        ga.worstInd()->printOn(cout);
        cout << endl;
        cout << "Online : " << ga.objOnLine() << endl;
        cout << "Offline : " << ga.objOffLine() << endl;
}

