// $Header: /home/anthony/tolkien/include/RCS/slctrw.h,v 1.4 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined(__SLCTRW_H)
#define __SLCTRW_H

#include <populatn.h>
#include <ranksb.h>

template <class Collection = vector<TGAObj*> >
class  RW_Select : public SelectionScheme<Collection>
{
//
//      Roulette Wheel Selection Scheme
//
public:
       RW_Select() : SelectionScheme<Collection>() { }
       ~RW_Select() {}

       void init()
       {
          // generate a random sequence of [0 .. N-1]
          cumSlots.reserve(pPop->popSize() + 1); // sumFitness() needs this size
          cumSlots.erase(cumSlots.begin(), cumSlots.end());
          ranksb.random(ndx, pPop->popSize(), pPop->popSize());
          sumFitness();
       }

       const TGAObj* select()
       {
	     count();
             return (const TGAObj*) *(pPop->begin() + _select());
       }

protected:
       double   flSumFitness;
       vector<double>cumSlots; // store the cumluative totals
       vector<int> ndx;   // ndx is a random sequence of
                          // the set of integers [0 .. N-1]
	                  // this is used to cancel
                          // positional bias in the
                          // population to be selected

       long _select()
       {
           double flRand = tRand.asDouble() * flSumFitness;
           long first = 1, last = pPop->popSize(), index;

           while (first <= last) {
               index = (first + last) / 2;
               // return index if flRand is smaller than or equal to
               // the cumulative total at index and larger than the
               // cumulative total at (index - 1)
               if ((flRand <= cumSlots[index]) && (flRand > cumSlots[index - 1]))
                   return ndx[index - 1];
               else if (flRand < cumSlots[index])
                        last = index - 1;
               else
                        first = index + 1;
           }

           // should not reach here
       }

       void sumFitness()
       {
          Collection::const_iterator front, cur;
          long i,j;

          cumSlots.push_back(0);
          for (front=pPop->begin(), i = 0, flSumFitness = 0; 
               i != pPop->popSize(); ++i) {
              cur = front + ndx[i];
              flSumFitness += (*cur)->fitness();
              cumSlots.push_back(flSumFitness);
          }
       }
};

#endif

