// $Header: /home/anthony/tolkien/include/RCS/slctrwwr.h,v 1.3 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined(__SLCTRWWR_H)
#define __SLCTRWWR_H

// #include <select.h>
#include <populatn.h>
#include <list.h>
#include <ranksb.h>

template <class Collection = vector<TGAObj*> >
class  RWwoR_Select : public SelectionScheme<Collection>
{
//
//      Roulette Wheel Selection Scheme without replacement
//
public:

	RWwoR_Select() : SelectionScheme<Collection>(), curSz(0) { }

        void init()
        {
            curSz = pPop->popSize();
            buildList();
            sumFitness();
        }

	const TGAObj* select()
        {
            double flPartSum;
            list<TGAObj*>::iterator p;
            double flRand;
            TGAObj* pObj;

            if (curSz <= 0)
                init();

            count();

            if (curSz > 0) {
                flRand = tRand.asDouble() * flSumFitness;
                p = objlist.begin();
                pObj = *p;
                flPartSum = pObj->fitness();
                if ((flRand > flPartSum) || (curSz == 1)) {
                    objlist.erase(p);
                    flSumFitness -= pObj->fitness();
                    curSz--;
                    return pObj;
                }
                else {
                    do {
                        ++p;
                        flPartSum += (*p)->fitness();
                    } while ((flPartSum < flRand) && (p != objlist.end()));
                }

                pObj = *p;
                objlist.erase(p);
                flSumFitness -= pObj->fitness();
                curSz--;
                return pObj;
            }

            // should not reach here
            return (const TGAObj*) NULL;
        }


protected:
    list<TGAObj*>   objlist;
    long            curSz;
    double          flSumFitness;

    void  buildList()
    {
          objlist.erase(objlist.begin(), objlist.end());

          vector<int> ndx(pPop->popSize());
          ranksb.random(ndx,pPop->popSize(),pPop->popSize());
          // ndx is used to prevent positional bias

          for (Collection::const_iterator i = pPop->begin();
               i != pPop->end(); ++i) 
                objlist.push_back(*i);
    }

    void  sumFitness()
    {
        Collection::const_iterator i;
        for (i = pPop->begin(), flSumFitness = 0; i != pPop->end(); ++i) 
             flSumFitness += (*i)->fitness();
    }

};

#endif

