// $Header: /home/anthony/tolkien/include/RCS/slctsus.h,v 1.5 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//
#if !defined(__SLCTSUS_H)
#define __SLCTSUS_H

#include <populatn.h>
#include <stack.h>

//
//      Stochastic Universal Sampling
//

template <class Collection = vector<TGAObj*> >
class SUS_Select : public SelectionScheme<Collection>
//
// Stochastic Universal Sampling
//
//	Baker, J.E.  Reducing Bias and Inefficency in the Selection Algorithm.
//		     Genetic Algorithms and the Applications :
//			Proceedings of the 2nd International Conference on 
//			Genetic Algorithms
//
{
public:
      SUS_Select() : SelectionScheme<Collection>() { }

      ~SUS_Select()  { }

      void selectOn(PopulationBase<Collection>& src)
      {
          pPop = (PopulationBase<Collection>*) &src;
          init();      
      }

      void init()
      {
          int i;
          double sum, factor, ptr;                          
          Collection::const_iterator begin = pPop->begin();
          vector<double> expVal;

          ndx.erase(ndx.begin(), ndx.end());
          ndx.reserve(pPop->popSize());
          expVal.reserve(pPop->popSize());

          // generate a random sequence of [0 .. N-1]
          // individuals are accessed using this sequence
          // to reduce position bias
          ranksb.random(ndx, pPop->popSize(), pPop->popSize());

          // calculate sum of fitness
          for (i = 0, factor = 0; i < pPop->popSize(); ++i)
               factor += (*(begin + i))->fitness();

          // calculate expected offsprings for each individual
          factor = pPop->popSize() / factor;
          for (i = 0; i < pPop->popSize(); ++i)
               expVal.push_back( (*(begin + ndx[i]))->fitness() * factor );

          ptr = tRand.asDouble();
          // the SUS algorithm starts here
          for (sum=i=0; i < pPop->popSize(); ++i)
               for (sum += expVal[i]; sum > ptr; ptr += 1)
                    mates.push( *(begin + ndx[i]) );
       }

	const TGAObj* select()
        {
            const TGAObj* pObj;
            if (mates.empty())
                init();
            pObj = mates.top();
            mates.pop();
            count();
            return pObj;
        }

protected :
	stack< vector<TGAObj*> > mates;
        vector<int> ndx; // ndx is a random sequence of
                         // the set of integers [0 .. N-1]
                         // this is used to cancel
                         // positional bias in the
                         // population to be selected

};

#endif

