// $Header: /home/anthony/tolkien/include/RCS/tbitstr.h,v 1.3 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined( __TBITSTR_H )
#define __TBITSTR_H

#include <BitString.h>

//
// some new functions on BitString
//
BitString toGray(const BitString &);       // convert binary to gray
BitString deGray(const BitString &);       // convert gray to binary

BitString BStrOnePtCrossover(const BitString&, const BitString&);
BitString BStrTwoPtCrossover(const BitString&, const BitString&);
BitString BStrOddPtCrossover(const BitString&, const BitString&, int);
BitString BStrEvenPtCrossover(const BitString&, const BitString&, int);
BitString BStrUniformCrossover(const BitString&, const BitString&, double);
long double bin2double(const BitString &, int, int);
unsigned long  bin2long(BitString, int, int);

//
// class TBitString is designed since some functions can be made more
// efficient if access to the internal of class BitString is granted
//
class TBitString : public BitString
{
public:
// constructors
    TBitString();
    TBitString(const BitString&);
    TBitString(const BitString&, int, int);
    TBitString(const BitSubString& y);

    unsigned long   asLong() const;
    long double     asDouble() const;
    void	    resize(int);

    void operator=(const BitString &);

};


// constructors & assignment

inline TBitString::TBitString() : BitString() {}

inline TBitString::TBitString(const BitString& x, int from, int length) :
       BitString(((BitString &) x).at(from, length))
{
}

inline TBitString::TBitString(const BitSubString& y) : BitString(y) {}

inline TBitString::TBitString(const BitString& y) : BitString(y) {}

inline void TBitString::resize(int len)
{
	rep = BStr_resize(rep, len);	
}

inline  void TBitString::operator=(const BitString &y)
{
	BitString::operator=(y);
}

inline	unsigned long TBitString::asLong() const
//
//	convert BitStrRep to an unsigned long
//      assuming that _BS_word is unsigned long (g++/bitprims.h)
{
     return * ((const _BS_word *) rep->s);
}

#endif

