// $Header: /home/anthony/tolkien/include/RCS/tritstr.h,v 1.4 1996/06/19 02:19:53 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#if !defined ( __TRITSTR_H )
#define   __TRITSTR_H

#include <tgaobj.h>
#include <trand.h>
#include <BitString.h>

class	Trit;

class TritSubString
{
  friend class TritString;

protected:

	TritString & tstr;
	int	     pos;
	int	     len;

	TritSubString(TritString& x, int p, int l);
	TritSubString(const TritString & x, int p, int l);
	TritSubString(const TritSubString& x);

public:

  void  operator= (const TritString &);
  void  operator= (const TritSubString&);

  int   length() const;
  int   empty() const;
};

class   TritString 
{
//      a string of ternary digit
public:
    TritString(double mutDC = DEFAULT_MUTDC);
    TritString(const BitString &, double mutDC = DEFAULT_MUTDC);
    TritString(const TritString &);
    TritString(const BitString &, 
               const BitString &, double mutDC = DEFAULT_MUTDC);
    TritString(const TritSubString &);
    TritString(const char *, double mutDC = DEFAULT_MUTDC);
    TritString(unsigned, bool=false, double mutDC = DEFAULT_MUTDC);

    void     printOn( ostream  & ) const;
    unsigned length() const;

    void     setRand(unsigned);
    void     randomize();
    bool     flip(double) const;
    unsigned mutate(double);

    void operator=(const TritString &);
    void operator=(const TritSubString &);
    void operator=(const BitString &);
    void operator=(const char *);
    void operator=(unsigned int);

    void operator+= (const TritString &);
    bool matches(const BitString &) const;
    bool contains(const BitString &) const;
    unsigned      findDontCare() const;
    void  clear();
    void  set();
    Trit  operator[](unsigned);
    void  resize(unsigned);
    void           assign(unsigned, const TritString &);
    void           setAt(unsigned,int);
    int            valueAt(unsigned) const;
    TritSubString  at(int, int) const;
    TritSubString  before(int) const;
    BitString      filter(const BitString &) const;

    void           onePtCrossover(const TritString &, const TritString &);
    void           twoPtCrossover(const TritString &, const TritString &);
    void           oddPtCrossover(const TritString &, const TritString &, int);
    void           evenPtCrossover(const TritString &, const TritString &, int);
    void           uniformCrossover(const TritString &, const TritString &, double);

    double         asDouble(int,int) const;
    unsigned long  asLong(int,int) const;
    double         asDouble() const;
    unsigned long  asLong() const;
    BitString      asBitString() const;

    unsigned matchCount(const TritString & src) const;

    double mutateDC() const;
    void  mutateDC(double);

    // friends
    friend TritString shorttoTritString(unsigned short);
    friend TritString longtoTritString(unsigned long);

    friend bool operator == (const TritString &, const TritString &);
    friend bool operator != (const TritString &, const TritString &);
    friend bool operator < (const TritString &, const TritString &);
    friend TritString  operator + (const TritString &, const TritString &);
    friend class TritSubString;
    friend class Trit;

protected:

    BitString  bits;
    BitString  dcbits;
    double     flMutDC;
};

inline int TritString::valueAt(unsigned uAt) const
{
	return dcbits.test(uAt) ?
		(int) dontCare : bits.test(uAt);
}

class   Trit 	// a ternary digit; 0=0, 1=1, -1=#
{
public:
	Trit(TritString&, unsigned);
	Trit(const Trit & );
	Trit& operator=(int);
	operator int();

private:
	TritString&    tstring;
	unsigned       uOffset;
};


inline  Trit::Trit(TritString& tstr, unsigned uAt) :
	tstring(tstr), uOffset(uAt)
{
}

inline  Trit::Trit(const Trit & from) :
	tstring(from.tstring), uOffset(from.uOffset)
{
}

inline  Trit::operator int()
{
	return tstring.valueAt(uOffset);
}

inline TritSubString::TritSubString(const TritSubString& x) :
	tstr(x.tstr), pos(x.pos), len(x.len)
{
}

inline TritSubString::TritSubString(TritString& x, int p, int l) :
	tstr(x), pos(p), len(l)
{
}

inline TritSubString::TritSubString(const TritString& x, int p, int l) :
	tstr((TritString&) x), pos(p), len(l)
{
}

inline  int  TritSubString::length() const
{
	return len;
}

inline  int  TritSubString::empty() const
{
	return len == 0;
}

inline TritString::TritString(double mutDC) : flMutDC(mutDC)
{
}

inline TritString::TritString(const BitString & bstr, double mutDC) :
                bits(bstr), dcbits(bstr), flMutDC(mutDC)
{
	dcbits.clear();
}

inline TritString::TritString(const TritString & src) :
                bits(src.bits), dcbits(src.dcbits), flMutDC(src.flMutDC)
{
}

inline TritString::TritString(const BitString & x, 
                              const BitString & y, 
                              double mutDC) :
        bits(x), dcbits(y), flMutDC(mutDC)
{
}

inline  Trit TritString::operator[](unsigned uAt)
{
	return Trit(*this, uAt);
}

inline unsigned TritString::length() const
{
    return bits.length();
}

inline void TritString::operator=(unsigned int value)
{
    *this = shorttoTritString(value);
}

inline unsigned TritString::findDontCare() const
{
	return dcbits.count();
}

inline void TritString::clear()
{
    bits.clear();
    dcbits.clear();
}

inline void TritString::set()
{
    bits.set();
    dcbits.set();
}

inline void TritString::assign(unsigned nAt, const TritString & src)
{
    bits.assign(nAt, src.bits.test(nAt));
    dcbits.assign(nAt, src.dcbits.test(nAt));
}

inline double TritString::mutateDC() const
{
	return flMutDC;
}

inline void TritString::mutateDC(double newMutDC)
{
	flMutDC = rangeCheck(newMutDC, 0.0, 1.0, DEFAULT_MUTDC);
}

inline unsigned long TritString::asLong(int nFrom, int nLength) const
{
        return (unsigned long) asDouble(nFrom, nLength);
}

inline double TritString::asDouble() const
{
        return asDouble(0, length());
}

inline unsigned long TritString::asLong() const
{
        return (unsigned long) asDouble(0, length());
}

inline bool TritString::flip(double flBias) const
{
	return tRand.flip(flBias);
}

#endif

