// $Header: /home/anthony/tolkien/include/RCS/tritstr1.h,v 1.3 1996/06/19 02:19:53 anthony Exp $                                              
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//                                                                       
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//   
#if !defined(__TRITSTR1_H)
#define __TRITSTR1_H

#include <BitString.h>
#include <charhpld.h>
#include <chromo.h>

//
// implementation of tri-ary strings using vector<char>
//
class TritString : public CharHaploid
{
public:                            
        TritString() { }

        TritString(int n, char c = 0) : CharHaploid(n,c), 
                                         flMutDC(0) {}

        TritString(unsigned len, bool randFlag = false, 
                    double mutDC = DEFAULT_MUTDC) :
               CharHaploid(len,0), flMutDC(mutDC)
        {
               if (randFlag)
                   randomize();
        }

        void resize(int newLen) 
        {
             int diff = newLen - length();
             if (diff >= 0) {
                genes.reserve(newLen);
                while (diff-- > 0)
                   genes.push_back(0);
             }
             else { // shrink
                iterator i = end();
                i += diff;
                genes.erase(i,end());
             }
        }

        TritString* clone() const
        {
             return new TritString(*this);
        }

        bool flip(double bias) const
        {
             return tRand.flip(bias);
        }

        double mutateDC() const
        {
            return flMutDC;
        }

        void mutateDC(double newMutDC)
        {
            flMutDC = rangeCheck(newMutDC, 0.0, 1.0, DEFAULT_MUTDC);
        }

        int valueAt(unsigned pos)
        { 
            return genes[pos] == (char) dontCare ? -1 : genes[pos];
        }

        void setAt(unsigned pos, int val)
        {
            if (val == (int) dontCare)
                genes[pos] = (char) dontCare;
            else
                genes[pos] = (val > 0) ? 1 : 0 ;
        }

        void setRand(unsigned pos);
        void randomize();
        void mutate(double);
        void printOn(ostream& out) const;

        void operator=(const BitString&);

        void oddPtCrossover(const TritString& t1, 
                            const TritString& t2, int nXpts)
        {
             ::oddPtCrossover(t1.genes,t2.genes,genes,nXpts);
        }

        void evenPtCrossover(const TritString& t1, 
                             const TritString& t2, int nXpts)
        {
             ::evenPtCrossover(t1.genes,t2.genes,genes,nXpts);
        }

         
        void uniformCrossover(const TritString& t1, 
                              const TritString& t2, double xrate)
        {
             ::uniformCrossover(t1.genes,t2.genes,genes,xrate);
        }

        void onePtCrossover(const TritString& t1, 
                            const TritString& t2) 
        {
             oddPtCrossover(t1,t2,1);
        }

        void twoPtCrossover(const TritString& t1, 
                            const TritString& t2) 
        {
             evenPtCrossover(t1,t2,2);
        }

protected:
        double     flMutDC;
};

#endif

