// $Header: /home/anthony/tolkien/src/RCS/dpldbin.cc,v 1.4 1996/06/19 02:24:55 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#include <dpldbin.h>
#include <trand.h>
#include <xover.h>

BinDiploid::BinDiploid(const TritString & src1,
                       const TritString & src2,
                       bool  grayFlag)
{
	hmlg1 = src1;
        hmlg2 = src2;
        hmlg1.mutateDC(0.5);
        hmlg2.mutateDC(0.5);
	normalize();
	setDominance();
        expressed.isGrayCode(grayFlag);
}

BinDiploid::BinDiploid(unsigned nLength,
                       bool uFlag,
                       bool grayFlag) :
	    hmlg1(nLength, uFlag, 0.5),
	    hmlg2(nLength, uFlag, 0.5)
{
        expressed.resize(hmlg1.length());
	setDominance();
        expressed.isGrayCode(grayFlag);
}

void    BinDiploid::normalize()
{
	if (hmlg1.length() > hmlg2.length())
	    hmlg2.resize(hmlg1.length());
	else if (hmlg2.length() > hmlg1.length())
	    hmlg1.resize(hmlg2.length());

	expressed.resize(hmlg1.length());
}

void BinDiploid::setDominance()
{
      for (register int nI = expressed.length() - 1; nI >= 0; nI--)
	   mapDominanceAt(nI);
}

void BinDiploid::printOn( ostream  & out) const
{
	    hmlg1.printOn(out);
	    out << " : ";
	    hmlg2.printOn(out);
	    out << " => ";
	    expressed.asBitString().printon(out);
	    out << ' ';
            TGAObj::printOn(out);
}

void BinDiploid::operator=(const BitString & bits)
{
        hmlg1 = bits;
        hmlg2 = hmlg1;
	normalize();
	setDominance();
}

void BinDiploid::operator=(const BinHaploid & source)
{
	hmlg1 = source.asBitString();
	hmlg2 = hmlg1;
	normalize();
	setDominance();
}

void BinDiploid::operator=(const BinDiploid & source)
{
	if (! isSame(source) ) {
	     hmlg1 = source.hmlg1;
	     hmlg2 = source.hmlg2;
	     normalize();
	     setDominance();
	}
}

void  BinDiploid::randomize()
{
	hmlg1.randomize();
	hmlg2.randomize();
	setDominance();
}

void  BinDiploid::mutate(double flRate)
{
	for (register int nI=length() - 1; nI >=0; nI--)
	     if (flip(flRate)) {
		 hmlg1.setRand(nI);
		 hmlg2.setRand(nI);
		 mapDominanceAt(nI);
	     }
}

void BinDiploid::resize(unsigned uNewLen)
{
    hmlg1.resize(uNewLen);
    hmlg2.resize(uNewLen);
    expressed.resize(uNewLen);
}

TGAObj*  BinDiploid::onePtCrossover(const BinIndividual* pObj) const
{
      return ((const BinIndividual *) pObj)->_onePtCrossover(this);
}

TGAObj*  BinDiploid::twoPtCrossover(const BinIndividual* pObj) const
{
      return ((const BinIndividual *) pObj)->_twoPtCrossover(this);
}

BinDiploid*  BinDiploid::oddPtCrossover(const TGAObj * pObj, int nXpts) const
{
      return (BinDiploid*) ((const BinIndividual *) pObj)->_oddPtCrossover(this, nXpts);
}

BinDiploid* BinDiploid::evenPtCrossover(const TGAObj * pObj, int nXpts) const
{
      return (BinDiploid*) ((const BinIndividual *) pObj)->_evenPtCrossover(this, nXpts);
}

BinDiploid* BinDiploid::uniformCrossover(const TGAObj * pObj, double flXRate) const
{
      return (BinDiploid*) ((const BinIndividual *) pObj)->_uniformCrossover(this, flXRate);
}

TGAObj* BinDiploid::_onePtCrossover(const BinDiploid * pMate) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.onePtCrossover(hmlg1, hmlg2);
   pChild->hmlg2.onePtCrossover(pMate->hmlg1, pMate->hmlg2);

   pChild->normalize();
   pChild->setDominance();

   return pChild;
}

TGAObj* BinDiploid::_twoPtCrossover(const BinDiploid * pMate) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.twoPtCrossover(hmlg1, hmlg2);
   pChild->hmlg2.twoPtCrossover(pMate->hmlg1, pMate->hmlg2);

   pChild->normalize();
   pChild->setDominance();

   return pChild;
}

TGAObj* BinDiploid::_onePtCrossover(const BinHaploid * pMate) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.onePtCrossover(hmlg1, hmlg2);
   pChild->hmlg2 = pMate->asBitString();

   pChild->normalize();
   pChild->setDominance();                                                    
   return pChild;
}

TGAObj* BinDiploid::_twoPtCrossover(const BinHaploid * pMate) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.twoPtCrossover(hmlg1, hmlg2);
   pChild->hmlg2 = pMate->asBitString();

   pChild->normalize();
   pChild->setDominance();
   return pChild;                                                             
}

TGAObj* BinDiploid::_oddPtCrossover(const BinDiploid * pMate, int nXpts) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.oddPtCrossover(hmlg1, hmlg2, nXpts);
   pChild->hmlg2.oddPtCrossover(pMate->hmlg1, pMate->hmlg2, nXpts);

   pChild->normalize();
   pChild->setDominance();

   return pChild;
}

TGAObj* BinDiploid::_oddPtCrossover(const BinHaploid * pMate, int nXpts) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.oddPtCrossover(hmlg1, hmlg2, nXpts);
   pChild->hmlg2 = pMate->asBitString();

   pChild->normalize();
   pChild->setDominance();
   return pChild;
}

TGAObj* BinDiploid::_evenPtCrossover(const BinDiploid * pMate, int nXpts) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.evenPtCrossover(hmlg1, hmlg2, nXpts);
   pChild->hmlg2.evenPtCrossover(pMate->hmlg1, pMate->hmlg2, nXpts);

   pChild->normalize();
   pChild->setDominance();
   return pChild;
}

TGAObj*  BinDiploid::_evenPtCrossover(const BinHaploid * pMate, int nXpts) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.evenPtCrossover(hmlg1, hmlg2, nXpts);
   pChild->hmlg2 = pMate->asBitString();

   pChild->normalize();
   pChild->setDominance();
   return pChild;
}

TGAObj* BinDiploid::_uniformCrossover(const BinDiploid * pMate, 
                                      double flXRate) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.uniformCrossover(hmlg1, hmlg2, flXRate);
   pChild->hmlg2.uniformCrossover(pMate->hmlg1, pMate->hmlg2, flXRate);

   pChild->normalize();
   pChild->setDominance();

   return pChild;
}

TGAObj* BinDiploid::_uniformCrossover(const BinHaploid * pMate, 
                                      double flXRate) const
{
   BinDiploid* pChild = (BinDiploid *) clone();

   pChild->hmlg1.uniformCrossover(hmlg1, hmlg2, flXRate);
   pChild->hmlg2 = pMate->asBitString();

   pChild->normalize();
   pChild->setDominance();
   return pChild;
}

void  BinDiploid::invert()
{
	int nVal, ndx;

	expressed.invert();
        for (ndx = length() - 1; ndx > 0; ndx--) {
	     if ((nVal = hmlg1.valueAt(ndx) + 1) > 1)
	         nVal = dontCare;
             hmlg1.setAt(ndx, nVal);
	     if ((nVal = hmlg2.valueAt(ndx) + 1) > 1)
	         nVal = dontCare;
	     hmlg2.setAt(ndx, nVal);
        }
}

void  BinDiploid::invert(int nIndex)
{
	// add 1 to each allele
	// 1 + 1 -> -1, 0 + 1 -> 1, -1 + 1 -> 0
	int nVal;

	expressed.invert(nIndex);
	if ((nVal = hmlg1.valueAt(nIndex) + 1) > 1)
	    nVal = dontCare;
	hmlg1.setAt(nIndex, nVal);
	if ((nVal = hmlg2.valueAt(nIndex) + 1) > 1)
	    nVal = dontCare;
	hmlg2.setAt(nIndex, nVal);
}

void  BinDiploid::set()
{
	TritString *pH1, *pH2;
        int ndx;

	expressed.set();
        for (ndx = length() - 1; ndx > 0; ndx--) {
            if (flip(0.5)) {
                pH1 = & hmlg1;
                pH2 = & hmlg2;          
            }
            else {
                pH1 = & hmlg2;
                pH2 = & hmlg1;
            }
	    pH1->setAt(ndx,1);
            if (flip(0.5))
                pH2->setAt(ndx, 0);
            else
                pH2->setAt(ndx, dontCare);
        }
}

void  BinDiploid::set(int nIndex)
{
	TritString *pH1, *pH2;

	if (flip(0.5)) {
	    pH1 = & hmlg1;
	    pH2 = & hmlg2;
	}
	else {
	    pH1 = & hmlg2;
	    pH2 = & hmlg1;
	}

	pH1->setAt(nIndex,1);
	if (flip(0.5))
	    pH2->setAt(nIndex, 0);
	else
	    pH2->setAt(nIndex, dontCare);

	expressed.set(nIndex);
}

void  BinDiploid::clear()
{
	TritString *pH1, *pH2;
        int ndx;

	expressed.clear();
        for (ndx = length() - 1; ndx > 0; ndx--) {
            if (flip(0.5)) {
                pH1 = & hmlg1;
                pH2 = & hmlg2;          
            }
            else {
                pH1 = & hmlg2;
                pH2 = & hmlg1;
            }
	    pH1->setAt(ndx,0);
	    if (flip(0.5))
	        pH2->setAt(ndx, 0);
            else
                pH2->setAt(ndx, dontCare);
        }
}

void  BinDiploid::clear(int nIndex)
{
	TritString *pH1, *pH2;

	if (flip(0.5)) {
	    pH1 = & hmlg1;
	    pH2 = & hmlg2;
	}
	else {
	    pH1 = & hmlg2;
	    pH2 = & hmlg1;
	}

	pH1->setAt(nIndex,0);
	if (flip(0.5))
	    pH2->setAt(nIndex, 0);
	else
	    pH2->setAt(nIndex, dontCare);

	expressed.clear(nIndex);
}

