// $Header: /home/anthony/tolkien/src/RCS/hpldbin.cc,v 1.6 1996/06/19 02:24:55 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//	Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//

#include <dpldbin.h>
#include <trand.h>
#include <xover.h>
#include <graycode.h>

BinHaploid::BinHaploid(unsigned uLen, bool uFlag, bool bFlag) :
                       graycodeFlag(bFlag) 
{
  resize(uLen);
  if (uFlag == true) 
      randomize(); 
}

BinHaploid::BinHaploid(const BitString & src, bool bFlag) : bits(src),
		       graycodeFlag(bFlag) { }

BinHaploid::BinHaploid(const char  *pString, bool bFlag) :
	bits(atoBitString(pString)), graycodeFlag(bFlag) { }

double  BinHaploid::asDouble(int nFrom, int nLength) const
{
        if (graycodeFlag) {
            TBitString bstr(bits, nFrom, nLength);

            // BitString bstr = bits.at(nFrom, nLength) is not used
            // because BitString::at() is not a const function

            if (bstr.length() <= LONGBITS) // use tGrayCode for long values
                return tGrayCode.asBin(bin2long(bstr,0,bstr.length()));
            else {
                bstr = deGray(bstr);
                bstr.right_trim(0);
                return bin2double(bstr, 0, bstr.length());
            }
        }
	else
            return bin2double(bits, nFrom, nLength);
}

unsigned long   BinHaploid::asLong(int nFrom, int nLength) const
{
        if (graycodeFlag) {
            TBitString bstr(bits, nFrom, nLength);
            // BitString bstr = bits.at(nFrom, nLength) is not used
            // because BitString::at() is not a const function

            return tGrayCode.asBin(bin2long(bstr, 0, bstr.length()));
        }
        else
	    return bin2long(bits, nFrom, nLength);
}

void BinHaploid::operator=(const BinHaploid & source)
{
	if (! isSame(source) ) {
	    TGAObj::operator=(source);
	    bits = source.bits;
	    graycodeFlag = source.graycodeFlag;
	}
}

void BinHaploid::operator=(const BinDiploid & source)
{
        TGAObj::operator=(source);
        bits = source.asBitString();
        isGrayCode(source.isGrayCode());
}

void BinHaploid::operator=(unsigned long lVal)
{
	unsigned lastLength = bits.length();
	if (graycodeFlag)
	    bits = longtoBitString(tGrayCode.asGray(lVal));
	else
	    bits = longtoBitString(lVal);
	resize(lastLength);
}

void    BinHaploid::randomize()
{
        for (register int nI = bits.length() - 1;nI >= 0; nI--)
             if (flip(0.5))
		 bits.set(nI);
	     else
		 bits.clear(nI);
}

void BinHaploid::printOn( ostream & out) const
{
	bits.printon(out);
        out << ' ';
        TGAObj::printOn(out);
}

void BinHaploid::mutate(double flRate)
{
     for (register int nI=length() - 1; nI >=0; nI--)
          if (flip(flRate))
	      invert(nI);
}

TGAObj*  BinHaploid::onePtCrossover(const BinIndividual* pMate) const
{
      BinHaploid*  pChild = (BinHaploid*) clone();
      pChild->bits = ::BStrOnePtCrossover(bits, pMate->asBitString());
      return pChild;
}

TGAObj*  BinHaploid::twoPtCrossover(const BinIndividual* pMate) const
{
      BinHaploid*  pChild = (BinHaploid*) clone();
      pChild->bits = ::BStrTwoPtCrossover(bits, pMate->asBitString());
      return pChild;
}

BinHaploid*  BinHaploid::oddPtCrossover(const TGAObj *pMate, int nXpt) const
{
      BinHaploid*  pChild = (BinHaploid*) clone();
      pChild->bits = ::BStrOddPtCrossover(bits,
                       ((const BinIndividual *) pMate)->asBitString(), nXpt);
      return pChild;
}

BinHaploid*  BinHaploid::evenPtCrossover(const TGAObj * pMate, int nXpts) const
{
    BinHaploid* pChild = (BinHaploid*) clone();
    pChild->bits = ::BStrEvenPtCrossover(bits,
                        ((const BinIndividual *) pMate)->asBitString(), nXpts);
    return pChild;
}

BinHaploid*  BinHaploid::uniformCrossover(const TGAObj * pMate, double flXRate) const
{
    BinHaploid* pChild = (BinHaploid*) clone();
    pChild->bits = ::BStrUniformCrossover(bits,
                                     ((const BinIndividual *) pMate)->asBitString(),
		     		     flXRate);
    return pChild;
}

unsigned BinHaploid::__hammingDistance(const BitString & src) const
{
	BitString  tmp;
	xor(bits, src, tmp);
	return tmp.count();
}

TGAObj* BinHaploid::_onePtCrossover(const BinDiploid * pObj) const
{
        return pObj->_onePtCrossover(this);
}

TGAObj* BinHaploid::_twoPtCrossover(const BinDiploid * pObj) const
{
        return pObj->_twoPtCrossover(this);
}

TGAObj* BinHaploid::_oddPtCrossover(const BinHaploid * pObj,  
                                    int nXpts) const
{
    BinHaploid*  pChild = (BinHaploid*) clone();
    pChild->bits = ::BStrOddPtCrossover(bits, pObj->bits,nXpts);
    return pChild;
}

TGAObj* BinHaploid::_evenPtCrossover(const BinHaploid * pObj, int nXpts) const
{
    BinHaploid*  pChild = (BinHaploid*) clone();
    pChild->bits = ::BStrEvenPtCrossover(bits, pObj->bits,nXpts);
    return pChild;
}

TGAObj* BinHaploid::_uniformCrossover(const BinHaploid * pObj, double flXRate) const
{
    BinHaploid*  pChild = (BinHaploid*) clone();
    pChild->bits = ::BStrUniformCrossover(bits, pObj->bits,flXRate);
    return pChild;
}

TGAObj* BinHaploid::_oddPtCrossover(const BinDiploid * pMate, int nXpts) const
{
        return pMate->_oddPtCrossover(this, nXpts);
}

TGAObj* BinHaploid::_evenPtCrossover(const BinDiploid * pMate, int nXpts) const
{
        return pMate->_evenPtCrossover(this, nXpts);
}

TGAObj* BinHaploid::_uniformCrossover(const BinDiploid * pMate,
                                             double flXRate) const
{
        return pMate->_uniformCrossover(this, flXRate);
}


