// $Header: /home/anthony/tolkien/src/RCS/ranksb.cc,v 1.3 1996/06/19 02:24:55 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
#include "ranksb.h"

RandIntSubSet    ranksb;

void RandIntSubSet::ordered(vector<int>& ints, int m, int n)
//
// use algorithm S and algorithm P of section 3.4.2 of
// Knuth's Seminumerical Algorithms to generate M
// random integers from 0, 1, .., N - 1
//
{
        int select, remaining, i, j;

	if (m < n) {
	   ints.reserve(m);
           ints.erase(ints.begin(),ints.end());
	   for (i=1, remaining = n, select = m;
		i <= n; i++, remaining--)
		if (tRand.asFloat() < ((float) select)/remaining) {
		    ints.push_back(i - 1);
		    select--;
		}
	}
	else { // generate a random permutation
	   ints.reserve(n);
           ints.erase(ints.begin(),ints.end());
	   for (i=0; i < n; i++)
		ints.push_back(i);
	   for (i=n - 1; i > 0; i--) {
		j = tRand(0,i);
		select = ints[j];
		ints[j] = ints[i];
		ints[i] = select;
	   }
	}
}

void RandIntSubSet::random(vector<int>& ints, int k, int n)
//
// generate K random integers from 0, 1, .., N - 1
// the elements of the subset are not sorted
//
{
        int r, i, m, j;

        if (k == n) {
            // shuffle the set [0 ... N-1]
            ints.erase(ints.begin(),ints.end());
            for (i = 0; i < n; i++)
                 ints.push_back(i);
            for (i = 0, r = n - 1; i < n; i++) {
                 j = tRand(i, r);  // swap ints[i] and ints[j]
                 m = ints[i];
                 ints[i] = ints[j];
                 ints[j] = m;
            }
        }
        else {
            vector<int> link(k,0);
            ints.reserve(k);
            fill_n(ints.begin(),k,-1);
            // ints[k] == -1 means element k is available

            for (r = k; r >= 0;) {
                 i = (m = (unsigned long) (tRand.asFloat() * n)) % k;

                 if (ints[i] < 0) { // slot available
                     ints[i] = m;
                     link[i] = 0;
                 }
                 else
                     if (m != ints[i]) {
                         while (link[i] > 0) {
                                if (ints[i] == m)
                                    break;
                                i = link[i];
                         }

                         if (ints[i] != m) {
                            // find home for m
                            while (--r > 0)
                                if (ints[r] < 0)
                                    break;
                            if (r >= 0) {
                                link[i] = r;
                                i = r;
                                ints[i] = m;
                                link[i] = 0;
                             }
                         }
                 }
            }
        }
}

