// $Header: /home/anthony/tolkien/src/RCS/tbitstr.cc,v 1.4 1996/06/19 02:24:55 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//

#include <math.h>
#include <tbitstr.h>
#include <ranksb.h>
#include <vector.h>

long double  TBitString::asDouble() const
//
// this function assumes that in the last _BS_word the bits
// not belonging to this bitstring is clean.
//
{
  int xwds = (unsigned) (rep->len) / BITSTRBITS + 1;
  int xlast = BitStr_pos(rep->len);
  const _BS_word * s = rep->s;
  const _BS_word * tops = &(s[xwds - 1]);

  long double accum = 0;
  int rdx = 0;
	
  while (s <= tops) {
     accum += ldexp(*s++, rdx);
     rdx += 32;
  }

  return accum;
}

BitString  toGray(const BitString & src) 	
{
	BitString       grayBits = src;
	register int i;
	register int last;

	for (i = src.length() - 1, last = 0; i >= 0; i--) {
	     grayBits.assign(i, src.test(i) != last);
	     last = src.test(i);
	}
	return grayBits;
}

BitString  deGray(const BitString & src)
{
	BitString       binaryBits = src;
	register int i;
	register int last;

	for (i = src.length() - 1, last = 0; i >= 0; i--) {
		if (src.test(i))
		    !last ? binaryBits.set(i) : binaryBits.clear(i);
		else
		    last ? binaryBits.set(i) : binaryBits.clear(i);
		last = binaryBits.test(i);
	}

	return binaryBits;
}

long double bin2double(const BitString & x, int nStartPos, int nLength) 
{
    if (nLength <= 0)
	return 0;

    register int nI, nJ;

    if (nStartPos < 0)
	nI = 0;
    else
	nI = (nStartPos > x.length() - 1) ? 0 : nStartPos;

    if (nLength < 0)
	nJ = 1;
    else
	nJ = (nLength > (nJ = x.length() - nI)) ? nJ : nLength;

    TBitString bstr = ((BitString &) x).at(nI, nJ);
    return bstr.asDouble();
}

unsigned long   bin2long(BitString x, int nStartPos, int nLength) 
{
    if (nLength <= 0)
	return 0;

    TBitString bstr(x, nStartPos, nLength);

    if (bstr.length() < LONGBITS)
	return bstr.asLong();
    else
	return LONG_MAX; 	// overflow
}

BitString  BStrEvenPtCrossover(const BitString& mother,
                               const BitString& father, int nXpts)
//
//      no checking on the length of the strings is performed
//      because this is assumed to be the task of the class Crossover
//
//      it is assumed that nXpts is an even number
//
{
        BitString offspring;
        int i, j, p, q, len, cont;
	bool flag;
        vector<int> xpts(nXpts);

	len = mother.length() - 1;
	offspring = mother;

	//
	// generate a sorted sequence of random numbers
	// in the range 0 to pShorter->length() - 1
	//
        j = nXpts - 1;
        do {
           cont = 0;         
           // since we treat the chromosome as a ring
           // 0 is equivalent to (length() - 1)
           // ensure the first point and the last point
           // do not overlap

           ranksb.ordered(xpts,nXpts,mother.length());

           for (i=nXpts-1; i > 0 && cont == 0; i--)
             if (xpts[i] == (xpts[i-1]+1))
                 cont = 1;
           if (xpts[0] == 0 && xpts[j] == len)
               cont = 1;
        } while ( cont == 1 );


        for (i = 0; i <= j; i+= 2) {
		p = xpts[i];
		q = xpts[i+1] - p + 1;
		offspring.at(p, q) = ((BitString&) father).at(p, q);
	}

	return offspring;
}

BitString  BStrOddPtCrossover(const BitString& mother,
		              const BitString& father, int nXpts)
//
//      no checking on the length of the strings is performed
//      because this is assumed to be the task of the class Crossover
//
//      so if length of father = length of mother = 1
//      a runtime error will occur
//
{
        BitString offspring = mother;
	int i, j, p, q, len;
	vector<int> xpts;

        //
	// generate a sorted sequence of random numbers
	// in the range 0 to length() - 2
	// and then map to 1 - (length() - 1)
        //
	ranksb.ordered(xpts, nXpts, father.length() - 1);

	for (i=0, j = nXpts - 1; i < j; i+=2) {
             p = xpts[i] + 1;
             q = xpts[i+1] + 1;
             len = q - p + 1;
             offspring.at(p,len) = ((BitString&) father).at(p,len);
	}

        // copy last segment
        p = xpts[nXpts - 1] + 1;
        len = father.length() - p;
        offspring.at(p,len) = ((BitString&) father).at(p,len);
	return offspring;
}

BitString  BStrUniformCrossover(const BitString& mother,
			        const BitString& father,
			        double flXRate) 
{
	BitString offspring(mother);

	for (register int nI = 0 ; nI < mother.length(); nI++)
	     if (tRand.flip(flXRate))
		 offspring.assign(nI, mother.test(nI));
	     else
		 offspring.assign(nI, father.test(nI));

	return offspring;
}

BitString  BStrOnePtCrossover(const BitString& mother,
		              const BitString& father)
//
//      no checking on the length of the strings is performed
//      because this is assumed to be the task of the class Crossover
//
//      so if length of father = length of mother = 1
//      a runtime error will occur
//
{
        BitString offspring = mother;
	int i, j, p, q, len;
        int xsite = tRand.asLong(1, father.length() - 1);
       
        len = mother.length() - xsite;
        offspring.at(xsite,len) = ((BitString&) father).at(xsite,len);

	return offspring;
}

BitString  BStrTwoPtCrossover(const BitString& mother,
                              const BitString& father)
//
//      no checking on the length of the strings is performed
//      because this is assumed to be the task of the class Crossover
//
//      it is assumed that nXpts is an even number
//
{
        BitString offspring;
        int xsite1, xsite2,diff,len = father.length() - 1;
	offspring = mother;

        do {
           // since we treat the chromosome as a ring
           // 0 is equivalent to (length() - 1)
           // ensure the two points do not overlap
           xsite1 = tRand.asLong(0,len);
           xsite2 = tRand.asLong(0,len);
           diff = abs(xsite1 - xsite2);
        } while (diff == 1 || diff == len);

        // ensure xsite1 < xsite2
        if (xsite1 > xsite2) {
            diff = xsite1;
            xsite1 = xsite2;
            xsite2 = diff;
        }
	diff = xsite2 - xsite1 + 1;
	offspring.at(xsite1, diff) = ((BitString&) father).at(xsite1, diff);

	return offspring;
}

