// $Header: /home/anthony/tolkien/src/RCS/tritstr.cc,v 1.6 1996/06/19 02:24:55 anthony Exp $
//
//      TOLKIEN: Tool Kit for Genetics-based Applications
//
//      Anthony Yiu-Cheung Tang, 1992-93.
//      Department of Computer Science
//      The Chinese University of Hong Kong.
//
//      Anthony Yiu-Cheung Tang, 1996.
//
#include   <tritstr.h>
#include   <tbitstr.h>
#include   <trand.h>
#include   <ranksb.h>
#include   <stdlib.h>

void TritString::resize(unsigned uNewSize)
{
        if (length() != uNewSize) {
            TBitString tbits = bits;
            tbits.resize(uNewSize);
            bits = tbits;
            tbits = dcbits;
            tbits.resize(uNewSize);
            dcbits = tbits;
        }
}

Trit& Trit::operator=(int nVal)
{
	tstring.setAt(uOffset, nVal);
	return *this;
}

TritString::TritString(const TritSubString & substr) : flMutDC(DEFAULT_MUTDC)
{
	bits = substr.tstr.bits.at(substr.pos,substr.len);
	dcbits = substr.tstr.dcbits.at(substr.pos,substr.len);
}

TritString::TritString(unsigned nLength, 
                       bool randFlag,
                       double mutDC) : flMutDC(mutDC)
{
	TBitString tbits;
        tbits.resize(nLength);
	bits = tbits;
	dcbits = tbits;

        if (randFlag)
            randomize();
	else {
	    bits.clear();
	    dcbits.clear();
	}
}

TritString::TritString(const char *szData, double mutDC) : flMutDC(mutDC)
{
	if (szData != NULL) {
	    register int nI = 0, nLength = strlen(szData);
            TBitString tbits;
	    tbits.resize(nLength);
            bits = tbits;
	    dcbits = tbits;

	    for (; nI < bits.length(); nI++)
		 switch (szData[nI]) {
		     case '0':
			bits.clear(nI);
			break;
		     case '1':
			bits.set(nI);
			break;
		     default:
			bits.set(nI);
			dcbits.set(nI);
			break;
		 }
	}
}

void    TritString::printOn(ostream & out) const
{
	for (register int nI = 0; nI < length(); nI++)
	     if (dcbits.test(nI))
		 out << (char) dontCare;
	     else
		 out << (unsigned) bits.test(nI);
}

unsigned  TritString::mutate(double flMRate)
{
	register int i,j;
	for (i=length()-1,j=0; i>=0; i--)
	     if (flip(flMRate)) {
		j++;
		setRand(i);
	     }
	return j;
}

void     TritString::randomize()
{
	register int i;
        for (i=length()-1;i>=0;i--) 
             setRand(i);
}

void    TritString::setAt(unsigned uAt, int nVal)
{
	switch(nVal) {
		case 0:
		   dcbits.clear(uAt);
		   bits.clear(uAt);
		   break;
		case 1:
		   bits.set(uAt);
		   dcbits.clear(uAt);
		   break;
		default:
		   bits.set(uAt);
		   dcbits.set(uAt);
		   break;
	}
}

void TritString::setRand(unsigned uAt)
{
        if (flMutDC == 0)
	    setAt(uAt, ((dcbits[uAt] ? -1 : bits.test(uAt)) +
                       (tRand(1,2) + 1)) % 3 - 1);
        else
            if (flip(flMutDC)) {
                // set to #
                dcbits.set(uAt);
                bits.set(uAt);
            }
            else if (flip(0.5)) {
               // set to 0
               dcbits.clear(uAt);
               bits.clear(uAt);
            }
            else {
               // set to 1
               bits.set(uAt);
               dcbits.clear(uAt);
            }
}

void TritString::operator=(const char * szData)
{
	if (szData != NULL) {
	    register int nI = 0, nLength = strlen(szData);
            TBitString tbits;
            tbits.resize(nLength);
	    bits = tbits;
	    dcbits = tbits;

	    for (; nI < bits.length(); nI++)
		 switch (szData[nI]) {
		     case '0':
			bits.clear(nI);
			break;
		     case '1':
			bits.set(nI);
			break;
		     default:
			bits.set(nI);
			dcbits.set(nI);
			break;
		 }
	}
}

void TritSubString::operator = (const TritString & str)
{
	tstr.bits.at(pos,len) = str.bits;
	tstr.dcbits.at(pos,len) = str.dcbits;
}

void TritSubString::operator = (const TritSubString& src)
{
	tstr.bits.at(pos,len) = src.tstr.bits.at(pos,len);
	tstr.dcbits.at(pos,len) = src.tstr.dcbits.at(pos,len);
}

void TritString::operator=(const TritSubString & y)
{
	bits = y.tstr.bits.at(y.pos,y.len);
	dcbits = y.tstr.dcbits.at(y.pos,y.len);
}

TritSubString TritString::at(int pos, int len) const
{
	return TritSubString(*this, pos, len);
}

TritSubString TritString::before(int pos) const
{
	return TritSubString(*this, 0, pos);
}

TritString  operator + (const TritString & x, const TritString & y)
{
	TritString r(x.bits + y.bits, x.dcbits + y.dcbits);
	return r;
}

bool    TritString::matches(const BitString & test) const
{
        return bits == (test | dcbits) ? true : false;
}

bool    TritString::contains(const BitString & test) const
{
	//
	// if the message is shorter than the TritString,
	// the function will return true if the bits of the
	// message match the corresponding bits of the classifier
	// part, e.g.
	//
	//      1000##000 will match the message 100
	//      1##0001   will not match the message 10110
	//
        return bits.contains(test | dcbits) ? true : false;
}

BitString    TritString::filter(const BitString & src) const
//
//      creates a bitstring using this tritstring as a template
//      and fills the don't cares with the corresponding bits
//      from src
//
{
	return bits & ~dcbits | src & dcbits;
}

void    TritString::operator += (const TritString & src)
{
	bits += src.bits;
	dcbits += src.dcbits;
}

void    TritString::onePtCrossover(const TritString & mother,
                                   const TritString & father)
{
        int i, j, p, q, len;
        int xsite = tRand.asLong(1, father.length() - 1);     
           
        bits = mother.bits;
        dcbits = mother.dcbits;

        len = mother.length() - xsite;         
        bits.at(xsite,len) = ((BitString&) father.bits).at(xsite,len);
        dcbits.at(xsite,len) = ((BitString&) father.dcbits).at(xsite,len);
}

void    TritString::twoPtCrossover(const TritString & mother,
                                   const TritString & father)
{
        int xsite1, xsite2,diff,len = father.length() - 1;

        bits = mother.bits;
        dcbits = mother.dcbits;

        do {
           // since we treat the chromosome as a ring
           // 0 is equivalent to (length() - 1)
           // ensure the two points do not overlap
           xsite1 = tRand.asLong(0,len);
           xsite2 = tRand.asLong(0,len);
           diff = abs(xsite1 - xsite2);
        } while (diff == 1 || diff == len);

        // ensure xsite1 < xsite2
        if (xsite1 > xsite2) {
            diff = xsite1;
            xsite1 = xsite2;
            xsite2 = diff;
        }
        diff = xsite2 - xsite1 + 1;
        bits.at(xsite1, diff) = ((BitString&) father.bits).at(xsite1, diff);
        dcbits.at(xsite1, diff) = ((BitString&) father.dcbits).at(xsite1, diff);
}

void    TritString::evenPtCrossover(const TritString & mother,
                                    const TritString & father, int nXpts)
//
//      no checking on the length of the strings is performed
//      because this is assumed to be the task of the class Crossover
//      it is assumed that nXpts is even
//
{
        int i, j, p, q, len, cont;                                   
        bool flag;                        
        vector<int> xpts(nXpts);

        len = mother.length() - 1;
        bits = mother.bits;
        dcbits = mother.dcbits;

        //
        // generate a sorted sequence of random numbers
        // in the range 0 to pShorter->length() - 1
        //
        j = nXpts - 1;
        do {
           cont = 0;
           // since we treat the chromosome as a ring
           // 0 is equivalent to (length() - 1)
           // ensure the first point and the last point
           // do not overlap

           ranksb.ordered(xpts,nXpts,mother.length());

           for (i=nXpts-1; i > 0 && cont == 0; i--)
             if (xpts[i] == (xpts[i-1]+1))
                 cont = 1;
           if (xpts[0] == 0 && xpts[j] == len)
               cont = 1;
        } while ( cont == 1 );


        for (i = 0; i <= j; i+= 2) {
                p = xpts[i];
                q = xpts[i+1] - p + 1;
                bits.at(p, q) = ((BitString&) father.bits).at(p, q);
                dcbits.at(p, q) = ((BitString&) father.dcbits).at(p, q);
        }
}

void    TritString::oddPtCrossover(const TritString& mother,
                                   const TritString& father, int nXpts)
//
//      no checking on the length of the strings is performed
//      because this is assumed to be the task of the class Crossover
//
{
        int i, j, p, q, len;
        vector<int> xpts;
        //
        // generate a sorted sequence of random numbers
        // in the range 0 to length() - 2
        // and then map to 1 - (length() - 1)
        //
        bits = mother.bits;
        dcbits = mother.dcbits;

        ranksb.ordered(xpts, nXpts, father.length() - 1);
        for (i=0, j = nXpts - 1; i < j; i+=2) {
             p = xpts[i] + 1;
             q = xpts[i+1] + 1;
             len = q - p + 1;
             bits.at(p,len) = ((BitString&) father.bits).at(p,len);
             dcbits.at(p,len) = ((BitString&) father.dcbits).at(p,len);
        }

        // copy last segment
        p = xpts[nXpts - 1] + 1;
        len = father.length() - p;
        bits.at(p,len) = ((BitString&) father.bits).at(p,len);
        dcbits.at(p,len) = ((BitString&) father.dcbits).at(p,len);
}

void	TritString::uniformCrossover(const TritString& mother,
				     const TritString& father,
                                     double flXRate)
{
	resize(mother.length());

	for (register int nI = 0 ; nI < mother.length(); nI++)
		if (flip(flXRate))
		    assign(nI, mother);
		else
		    assign(nI, father);
}

BitString TritString::asBitString() const
{
        BitString bstr(bits);
        bstr |= dcbits;
        return bstr;
}

unsigned TritString::matchCount(const TritString & src) const
{
//
//      count number of positions of similarity
//
        register int j, count;
        for (j = length() - 1, count = 0; j >= 0; j--)
                if (valueAt(j) == src.valueAt(j))
                    count++;
        return count;
}

TritString  shorttoTritString(unsigned short value)
//
//      create a TritString with the first position
//      as the least significant value
{
        BitString  bits, dcbits;
	int	quot, rem, s;

	quot = value / 3;
	rem  = value % 3;

        for (; quot > 3;) {
            switch( rem ) {
                case 0 :
                    bits += 0;
                    dcbits += 0;
                    break;
                case 1 :
		    bits += 1;
                    dcbits += 0;
                    break;
                case 2 :
                    bits += 1;
                    dcbits += 1;
                    break;
            }
	    s = quot;
	    quot =  s / 3;
	    rem  =  s % 3;
        }
        return TritString(bits, dcbits);
}

TritString  longtoTritString(unsigned long value)
{
        BitString  bits, dcbits;
        unsigned long quot, rem, s;

	quot = value / 3;
	rem  = value % 3;
	for (; quot > 3;) {
            switch( rem ) {
                case 0 :
                    bits += 0;
                    dcbits += 0;
                    break;
                case 1 :
                    bits += 1;
                    dcbits += 0;
                    break;
		case 2 :
                    bits += 1;
                    dcbits += 1;
                    break;
            }
	    s = quot;
	    quot = s / 3;
	    rem  = s % 3;
        }

        return TritString(bits, dcbits);
}

bool operator < (const TritString& t1, const TritString& t2)
{
	return t1.asDouble() < t2.asDouble();
}

bool operator == (const TritString& t1, const TritString& t2)
{
        return (t1.bits == t2.bits && t1.dcbits == t2.dcbits);
}

bool operator != (const TritString& t1, const TritString& t2)
{
        return ! operator==(t1,t2);
}

void TritString::operator=(const TritString & source)
{
	bits = source.bits;
	dcbits = source.dcbits;
	flMutDC = source.flMutDC;
}

void TritString::operator=(const BitString & source)
{
	bits = source;
	dcbits = source;
	dcbits.clear();
}


double TritString::asDouble(int from, int len) const
//
//      bits[i]         dcbits[i]       integer value
//      ---------------------------------------------
//        0                0                 0
//        1                0                 1
//        1                1                 2
//
{
    if (len <= 0) return 0;

    int i, j, k;
    double accum, power3 = 1;

    if (from < 0)
	i = 0;
    else
        i = (from > length() - 1) ? 0 : from;

    j = (len > (j = length() - i)) ? j : len;

    for (k = i; j > 0; k++, j--) {
	if ( bits.test(k) ) {
           if ( dcbits.test(k) ) // 11 -> 2
               accum += 2 * power3;
	   else // 10 -> 1
	       accum += power3;
	}
	power3 *= 3;
    }

    return accum;
}

